SamplesTensor2

Struct SamplesTensor2 

pub struct SamplesTensor2 {}
Holds second-order tensor samples

Implementations

Source

impl SamplesTensor2

pub const TENSOR_O: SampleTensor2

Collects data for a symmetric tensor with all zero components (Tensor O)

pub const TENSOR_I: SampleTensor2

Collects data for a symmetric diagonal tensor, the identity tensor (Tensor I)

pub const TENSOR_X: SampleTensor2

Collects data for a symmetric tensor in 2D (as in plane-stress analyses) (Tensor X)

pub const TENSOR_Y: SampleTensor2

Collects data for a symmetric tensor in 2D (as in plane-stress analyses) (Tensor Y)

pub const TENSOR_Z: SampleTensor2

Collects data for a symmetric tensor in 2D (Tensor Z)

pub const TENSOR_U: SampleTensor2

Collects data for a symmetric tensor in 3D (Tensor U)

pub const TENSOR_S: SampleTensor2

Collects data for a symmetric tensor in 3D (Tensor S)

pub const TENSOR_R: SampleTensor2

Collects data for a non-symmetric tensor in 3D (Tensor R)

pub const TENSOR_T: SampleTensor2

Collects data for a non-symmetric tensor in 3D (Tensor T)

pub const COAL_01: SampleTensor2

pub const COAL_12: SampleTensor2

pub fn all_symmetric<'a>() -> Vec<&'a SampleTensor2>