pub struct SamplesTensor4;Expand description
Holds fourth-order tensor samples
Implementations
Sourceimpl SamplesTensor4
impl SamplesTensor4
pub const SAMPLE1: [[[[f64; 3]; 3]; 3]; 3]
Sourcepub const SAMPLE1_STD_MATRIX: [[f64; 9]; 9]
pub const SAMPLE1_STD_MATRIX: [[f64; 9]; 9]
Sample matrix representation with standard components
Sourcepub const SAMPLE1_KELVIN_MATRIX: [[f64; 9]; 9]
pub const SAMPLE1_KELVIN_MATRIX: [[f64; 9]; 9]
Sample matrix representation with Kelvin-Mandel components
pub const SAMPLE2: [[[[f64; 3]; 3]; 3]; 3]
pub const SYM_SAMPLE1: [[[[f64; 3]; 3]; 3]; 3]
Sourcepub const SYM_SAMPLE1_STD_MATRIX: [[f64; 9]; 9]
pub const SYM_SAMPLE1_STD_MATRIX: [[f64; 9]; 9]
Sample matrix representation of symmetric tensor with standard components
Sourcepub const SYM_SAMPLE1_KELVIN_MATRIX: [[f64; 6]; 6]
pub const SYM_SAMPLE1_KELVIN_MATRIX: [[f64; 6]; 6]
Sample matrix representation of symmetric tensor with Kelvin-Mandel components
pub const SYM_2D_SAMPLE1: [[[[f64; 3]; 3]; 3]; 3]
Sourcepub const SYM_2D_SAMPLE1_STD_MATRIX: [[f64; 9]; 9]
pub const SYM_2D_SAMPLE1_STD_MATRIX: [[f64; 9]; 9]
Sample matrix representation for generalized plane spaces with standard components
Sourcepub const SYM_2D_SAMPLE1_KELVIN_MATRIX: [[f64; 4]; 4]
pub const SYM_2D_SAMPLE1_KELVIN_MATRIX: [[f64; 4]; 4]
Sample matrix representation for generalized plane spaces with Kelvin-Mandel components